import os
from pathlib import Path

import click
from robocop.config import ConfigManager, FileFiltersOptions

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    NOTE_MARKER,
    VERBOSE_DOUBLE,
    VERBOSE_NO,
    VERBOSE_SINGLE,
)


def cli_discover_file_paths(input_path: str, *, verbose: int) -> list[Path]:
    """
    Get file paths recursively with Robocop excludes.
    """
    click.echo(f"Discovering files in `{input_path}` using Robocop config...")

    robocop_config = ConfigManager(sources=[input_path])

    extensions = {"*.robot", "*.resource", "*.py"}
    if robocop_config.default_config.file_filters:
        robocop_config.default_config.file_filters.default_include = extensions
    else:
        robocop_config.default_config.file_filters = FileFiltersOptions(default_include=extensions)

    file_paths = [path[0] for path in robocop_config.paths]

    _log_file_stats(file_paths, input_path, verbose)
    return file_paths


def _log_file_stats(file_paths: list[Path], input_path: str, verbose: int) -> None:
    """
    Output details to the user
    """
    if len(file_paths) == 0:
        click.echo(f"{ERROR_MARKER} Found 0 files in `{input_path}`")
        click.echo(f"{NOTE_MARKER} All files in Robocop config `exclude` are ignored")
        click.echo(f"{NOTE_MARKER} All files listed in `.gitignore` files are ignored")

        if verbose < VERBOSE_DOUBLE:
            return

        click.echo(f"{NOTE_MARKER} All of the following files are excluded:")
        for root, _dirs, files in os.walk(input_path):
            for file in files:
                click.echo(f"{INDENT}{Path(root, file).as_posix()}")
        return

    if verbose == VERBOSE_NO:
        return

    click.echo(f"{DONE_MARKER} Discovered {len(file_paths)} files")

    if verbose == VERBOSE_SINGLE:
        return

    for path in file_paths:
        click.echo(INDENT + click.style(str(path), fg="bright_black"))



from .const import KeywordData


def sort_keywords_by_name(unsorted: list[KeywordData]) -> list[KeywordData]:
    """Sort list of keyword by library name and keyword name"""
    return sorted(unsorted, key=lambda kw: f"{kw.library or 'aaa'}.{kw.normalized_name}")



from pathlib import Path

import click

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    VERBOSE_NO,
    WARN_MARKER,
)
from robotframework_find_unused.common.visit import visit_robot_files
from robotframework_find_unused.visitors.library_import import LibraryData, LibraryImportVisitor


def cli_step_get_downloaded_lib_keywords(
    file_paths: list[Path],
    *,
    verbose: int,
    enrich_py_keywords: bool = False,
):
    """
    Gather keyword definitions from imported downloaded libraries and show progress

    Will only resolve libraries that are actually imported in an in-scope .robot or .resource file.
    """
    click.echo("Gathering downloaded library keyword definitions...")

    robot_file_paths = [p for p in file_paths if p.suffix in (".resource", ".robot")]

    visitor = LibraryImportVisitor(enrich_py_keywords=enrich_py_keywords)
    visit_robot_files(robot_file_paths, visitor)
    downloaded_library = list(visitor.downloaded_libraries.values())

    _log_downloaded_lib_stats(downloaded_library, verbose)
    return downloaded_library


def _log_downloaded_lib_stats(libraries: list[LibraryData], verbose: int) -> None:
    """
    Output details encountered downloaded libraries to the user
    """
    click.echo(
        (WARN_MARKER if len(libraries) == 0 else DONE_MARKER)
        + f" Found {len(libraries)} downloaded libraries",
    )

    if verbose == VERBOSE_NO:
        return

    for lib in libraries:
        if len(lib.keywords) == 0:
            # Import error
            click.echo(f"{INDENT}{lib.name}: {ERROR_MARKER}")
        else:
            click.echo(f"{INDENT}{lib.name}: {len(lib.keywords)} keywords")



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Literal, TypeAlias

import click
from robot.libdocpkg.model import ArgumentSpec

DONE_MARKER = "[ " + click.style("DONE", fg="green") + " ]"
WARN_MARKER = "[ " + click.style("WARNING", fg="yellow") + " ]"
ERROR_MARKER = "[ " + click.style("ERROR", fg="red") + " ]"
NOTE_MARKER = "[ " + click.style("NOTE", fg="cyan") + " ]"
INDENT = "    "
VERBOSE_NO = 0
VERBOSE_SINGLE = 1
VERBOSE_DOUBLE = 2

KeywordFilterOption: TypeAlias = Literal["include", "exclude", "only"]


@dataclass
class KeywordData:
    """Data structure for Keywords"""

    name: str
    normalized_name: str
    type: Literal[
        "CUSTOM_SUITE",
        "CUSTOM_LIBRARY",
        "CUSTOM_RESOURCE",
        "LIBRARY",
        "UNKNOWN",
    ]
    argument_use_count: None | dict[str, int]
    deprecated: None | bool
    private: bool
    use_count: int
    returns: None | bool
    """If True: Returns something. If False: Does not return anything. If None: Unknown"""

    return_use_count: int
    """If the keyword returns, how often is the return used during keyword call?"""

    arguments: ArgumentSpec | None
    library: str


VariableDefinedInType: TypeAlias = Literal["variables_section", "variable_file", "runtime"]


@dataclass
class VariableData:
    """Data structure for Variables"""

    name: str
    normalized_name: str
    resolved_name: str
    use_count: int
    defined_in_type: VariableDefinedInType
    defined_in: str
    value: Iterable[str]


@dataclass
class LibraryData:
    """Data structure for Library keywords"""

    name: str
    name_normalized: str
    keywords: list[KeywordData]
    keyword_names_normalized: set[str]



from pathlib import Path

import robot.api.parsing


def visit_robot_files(
    file_paths: list[Path],
    visitor: robot.api.parsing.ModelVisitor,
):
    """
    Use Robotframework to traverse files with a visitor.

    See Robotframework docs on Visitors for details.
    """
    for file_path in file_paths:
        model = robot.api.parsing.get_model(file_path, data_only=True)
        visitor.visit(model)



from pathlib import Path

import click
import robot.errors
from robot.libdoc import LibraryDocumentation
from robot.libdocpkg.model import LibraryDoc

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    INDENT,
    VERBOSE_NO,
    VERBOSE_SINGLE,
    WARN_MARKER,
)


def cli_step_parse_files(file_paths: list[Path], *, verbose: int):
    """
    Parse files with libdoc and keep the user up-to-date on progress
    """
    click.echo("Parsing files with LibDoc...")

    (parsed_files, errors) = _find_files_with_libdoc(file_paths)

    _log_file_errors(errors)
    _log_file_stats(parsed_files, verbose)
    return parsed_files


def _find_files_with_libdoc(file_paths: list[Path]) -> tuple[list[LibraryDoc], list[str]]:
    """
    Gather files in the given scope with LibDoc

    Libdoc supports .robot, .resource, .py, and downloaded libs
    """
    files: list[LibraryDoc] = []
    errors: list[str] = []
    for file in file_paths:
        try:
            libdoc = LibraryDocumentation(file)
            files.append(libdoc)
        except robot.errors.DataError as e:
            errors.append(e.message.split("\n", maxsplit=1)[0])
            continue
        if not isinstance(libdoc, LibraryDoc):
            continue

    return (files, errors)


def _log_file_errors(errors: list[str]) -> None:
    if len(errors) == 0:
        return

    click.echo(f"{WARN_MARKER} Failed to parse {len(errors)} files. Files will be ignored")

    for error in errors:
        click.echo(f"{INDENT}{WARN_MARKER} {error}")


def _log_file_stats(files: list[LibraryDoc], verbose: int) -> None:
    """
    Output details on parsed files to the user
    """
    click.echo(f"{DONE_MARKER} Parsed {len(files)} files")

    if verbose == VERBOSE_NO:
        return

    file_types: dict[str, list[str]] = {}
    for file in files:
        if not file.source:
            continue

        if file.type not in file_types:
            file_types[file.type] = []
        file_types[file.type].append(file.source)

    for file_type, file_paths in sorted(file_types.items(), key=lambda x: len(x[1]), reverse=True):
        click.echo(f"{INDENT}{len(file_paths)} files of type CUSTOM_{file_type}")

        if verbose == VERBOSE_SINGLE:
            continue
        for path in file_paths:
            click.echo(f"{INDENT}{INDENT}{click.style(path, fg='bright_black')}")



from pathlib import Path

import click

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    VERBOSE_NO,
    VariableData,
)
from robotframework_find_unused.common.visit import visit_robot_files
from robotframework_find_unused.visitors.variable_count import VariableCountVisitor


def cli_count_variable_uses(
    file_paths: list[Path],
    variable_defs: dict[str, VariableData],
    *,
    verbose: int,
):
    """
    Walk through all robot files to count keyword uses and show progress
    """
    click.echo("Counting variable usage...")
    variables = _count_variable_uses(file_paths, variable_defs)

    _log_variable_stats(variables, verbose)
    return variables


def _count_variable_uses(
    file_paths: list[Path],
    variables: dict[str, VariableData],
) -> list[VariableData]:
    """
    Walk through all robot files to count keyword uses.
    """
    visitor = VariableCountVisitor(variables)
    visit_robot_files(file_paths, visitor)

    return list(visitor.variables.values())


def _log_variable_stats(variables: list[VariableData], verbose: int) -> None:
    """
    Output details encountered downloaded libraries to the user
    """
    total_uses = 0
    for var in variables:
        total_uses += var.use_count
    click.echo(
        (ERROR_MARKER if total_uses == 0 else DONE_MARKER)
        + f" Found {total_uses} variable uses of gathered variables",
    )

    if verbose == VERBOSE_NO:
        return

    click.echo(f"{INDENT}Variable definitions")
    click.echo(f"{INDENT}{INDENT}Total\t{len(variables)}")

    unused_variables = [var.name for var in variables if var.use_count == 0]
    try:
        percentage_unused = round(len(unused_variables) / len(variables) * 100, 1)
    except ZeroDivisionError:
        percentage_unused = 0

    percentage_used = round((100 - percentage_unused), 1)
    click.echo(
        f"{INDENT}{INDENT}Used\t{len(variables) - len(unused_variables)}\t"
        + click.style(f"({percentage_used}%)", fg="bright_black"),
    )
    click.echo(
        f"{INDENT}{INDENT}Unused\t{len(unused_variables)}\t"
        + click.style(f"({percentage_unused}%)", fg="bright_black"),
    )

    click.echo(f"{INDENT}Variable usage")

    total_uses = 0
    for var in variables:
        total_uses += var.use_count
    click.echo(f"{INDENT}{INDENT}Total\t{total_uses} " + click.style("uses", fg="bright_black"))

    try:
        average = round(total_uses / len(variables), 1)
    except ZeroDivisionError:
        average = 0
    click.echo(
        f"{INDENT}{INDENT}Average\t{average} "
        + click.style("uses per gathered variable", fg="bright_black"),
    )



"""
Implementation of the 'arguments' command
"""

from dataclasses import dataclass

import click

from robotframework_find_unused.common.cli import cli_hard_exit, pretty_kw_name
from robotframework_find_unused.common.const import INDENT, KeywordData, KeywordFilterOption
from robotframework_find_unused.common.sort import sort_keywords_by_name

from .step.discover_files import cli_discover_file_paths
from .step.keyword_count_uses import cli_count_keyword_uses
from .step.keyword_definitions import cli_step_get_custom_keyword_definitions
from .step.keyword_filter import cli_filter_keywords
from .step.lib_keyword_definitions import cli_step_get_downloaded_lib_keywords
from .step.parse_files import cli_step_parse_files


@dataclass
class ArgumentsOptions:
    """
    Command line options for the 'arguments' command
    """

    deprecated_keywords: KeywordFilterOption
    private_keywords: KeywordFilterOption
    library_keywords: KeywordFilterOption
    unused_keywords: KeywordFilterOption
    keyword_filter_glob: str | None
    show_all_count: bool
    verbose: int
    source_path: str


def cli_arguments(options: ArgumentsOptions):
    """
    Entry point for the CLI command
    """
    file_paths = cli_discover_file_paths(options.source_path, verbose=options.verbose)
    if len(file_paths) == 0:
        return cli_hard_exit(options.verbose)

    files = cli_step_parse_files(
        file_paths,
        verbose=options.verbose,
    )

    keywords = cli_step_get_custom_keyword_definitions(
        files,
        verbose=options.verbose,
    )
    if len(keywords) == 0:
        return cli_hard_exit(options.verbose)

    downloaded_library_keywords = cli_step_get_downloaded_lib_keywords(
        file_paths,
        verbose=options.verbose,
    )

    counted_keywords = cli_count_keyword_uses(
        file_paths,
        keywords,
        downloaded_library_keywords=downloaded_library_keywords,
        verbose=options.verbose,
    )

    if options.library_keywords != "exclude" and options.unused_keywords != "exclude":
        for lib in downloaded_library_keywords:
            for kw in lib.keywords:
                if kw in counted_keywords:
                    continue
                counted_keywords.append(kw)

    counted_keywords = cli_filter_keywords(
        counted_keywords,
        filter_deprecated=options.deprecated_keywords,
        filter_private=options.private_keywords,
        filter_library=options.library_keywords,
        filter_unused=options.unused_keywords,
        filter_glob=options.keyword_filter_glob,
    )
    _cli_log_results(counted_keywords, options)
    return _exit_code(counted_keywords)


def _cli_log_results(keywords: list[KeywordData], options: ArgumentsOptions) -> None:
    click.echo()

    keywords = sort_keywords_by_name(keywords)

    for kw in keywords:
        if kw.argument_use_count is None:
            continue

        if options.show_all_count:
            _cli_log_results_show_count(kw)
        else:
            _cli_log_results_unused(kw)


def _cli_log_results_unused(kw: KeywordData) -> None:
    """
    Output a keywords arguments if they're unused
    """
    if not kw.arguments or len(kw.arguments.argument_names) == 0 or not kw.argument_use_count:
        return

    unused_args = {}
    for arg, count in kw.argument_use_count.items():
        if count == 0:
            unused_args[arg] = 0

    if not unused_args:
        return

    click.echo(pretty_kw_name(kw))

    click.echo(
        f"{INDENT}Unchanged arguments: {len(unused_args)} of {len(kw.argument_use_count)}",
    )
    for arg in unused_args:
        if arg in kw.arguments.defaults:
            click.echo(f"{INDENT}{INDENT}{arg}={kw.arguments.defaults[arg]}")
        else:
            click.echo(f"{INDENT}{INDENT}{arg}")

    click.echo()


def _cli_log_results_show_count(kw: KeywordData) -> None:
    """
    Output a keyword and all it's argument counts
    """
    arguments = kw.argument_use_count

    click.echo(pretty_kw_name(kw))

    if not arguments or len(arguments) == 0:
        click.echo(INDENT + click.style("Keyword has 0 arguments", fg="bright_black"))
        click.echo()
        return

    click.echo(f"{INDENT}use_count\targument")

    for arg, use_count in arguments.items():
        kw_args = kw.arguments
        if kw_args is not None and arg in kw_args.defaults:
            click.echo(f"{INDENT}{use_count}\t\t{arg}={kw_args.defaults[arg]}")
        else:
            click.echo(f"{INDENT}{use_count}\t\t{arg}")

    click.echo()


def _exit_code(keywords: list[KeywordData]) -> int:
    unused_args = 0
    for kw in keywords:
        if not kw.argument_use_count:
            continue

        for count in kw.argument_use_count.values():
            if count == 0:
                unused_args += 1

    exit_code = unused_args
    return min(exit_code, 200)



from pathlib import Path

import click

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    INDENT,
    VERBOSE_NO,
    WARN_MARKER,
    KeywordData,
    LibraryData,
)
from robotframework_find_unused.common.visit import visit_robot_files
from robotframework_find_unused.visitors.keyword_visitor import KeywordVisitor


def cli_count_keyword_uses(
    file_paths: list[Path],
    keywords: list[KeywordData],
    downloaded_library_keywords: list[LibraryData],
    *,
    verbose: int,
):
    """
    Walk through all robot files to count keyword uses and keep the user up-to-date on progress
    """
    click.echo("Counting keyword usage...")

    counted_keywords = _count_keyword_uses(
        file_paths,
        keywords,
        downloaded_library_keywords,
    )
    _log_keyword_call_stats(counted_keywords, verbose)

    unknown_keywords = [kw for kw in counted_keywords if kw.type == "UNKNOWN"]
    _log_unknown_keyword_stats(unknown_keywords, verbose)

    return counted_keywords


def _count_keyword_uses(
    file_paths: list[Path],
    keywords: list[KeywordData],
    downloaded_library_keywords: list[LibraryData],
) -> list[KeywordData]:
    """
    Walk through all robot files to count keyword uses.
    """
    visitor = KeywordVisitor(keywords, downloaded_library_keywords)
    visit_robot_files(file_paths, visitor)
    return list(visitor.keywords.values())


def _log_keyword_call_stats(keywords: list[KeywordData], verbose: int) -> None:
    """
    Output details on calls to the given keywords to the user
    """
    total_uses = sum([kw.use_count for kw in keywords])
    click.echo(
        (WARN_MARKER if total_uses == 0 else DONE_MARKER)
        + f" Processed {total_uses} keyword calls",
    )

    if verbose == VERBOSE_NO:
        return

    kw_type_use_count: dict[str, int] = {}
    for kw in keywords:
        if kw.type not in kw_type_use_count:
            kw_type_use_count[kw.type] = 0
        kw_type_use_count[kw.type] += kw.use_count

    for kw_type, count in sorted(kw_type_use_count.items(), key=lambda x: x[1], reverse=True):
        click.echo(f"{INDENT}{count} calls to keywords of type {kw_type}")

    click.echo(
        (WARN_MARKER if len(keywords) == 0 else DONE_MARKER)
        + f" Found {len(keywords)} unique keywords "
        + click.style("(keyword definitions and calls)", fg="bright_black"),
    )


def _log_unknown_keyword_stats(keywords: list[KeywordData], verbose: int) -> None:
    """
    Output details on keywords for which no definition was found
    """
    if len(keywords) > 0:
        click.echo(
            f"{WARN_MARKER} Found {len(keywords)} called keywords without a definition",
        )

    if verbose == VERBOSE_NO:
        return

    for kw in keywords:
        click.echo(f"{INDENT}{kw.name}")



import fnmatch
from collections.abc import Callable

import click

from robotframework_find_unused.common.const import NOTE_MARKER, KeywordData, KeywordFilterOption


def cli_filter_keywords(  # noqa: PLR0913
    keywords: list[KeywordData],
    *,
    filter_deprecated: KeywordFilterOption | None = None,
    filter_private: KeywordFilterOption | None = None,
    filter_library: KeywordFilterOption | None = None,
    filter_unused: KeywordFilterOption | None = None,
    filter_returns: KeywordFilterOption | None = None,
    filter_glob: str | None,
) -> list[KeywordData]:
    """
    Filter a list of keywords according to the user options.

    Logs to the user which type of keywords are excluded.

    Returns a filtered list.
    """
    if filter_deprecated:
        keywords = _cli_filter_keywords_by_option(
            keywords,
            filter_deprecated,
            lambda kw: kw.deprecated or False,
            "deprecated",
        )

    if filter_private:
        keywords = _cli_filter_keywords_by_option(
            keywords,
            filter_private,
            lambda kw: kw.private,
            "private",
        )

    if filter_library:
        keywords = _cli_filter_keywords_by_option(
            keywords,
            filter_library,
            lambda kw: kw.type == "LIBRARY",
            "downloaded library",
        )

    if filter_unused:
        keywords = _cli_filter_keywords_by_option(
            keywords,
            filter_unused,
            lambda kw: kw.use_count == 0,
            "unused",
        )

    if filter_returns:
        keywords = _cli_filter_keywords_by_option(
            keywords,
            filter_returns,
            lambda kw: kw.returns is True,
            "returning",
        )

    if filter_glob:
        click.echo(f"Only showing keywords matching '{filter_glob}'")

        pattern = filter_glob.lower()
        keywords = list(
            filter(
                lambda kw: fnmatch.fnmatchcase(kw.name.lower(), pattern),
                keywords,
            ),
        )

    return keywords


def _cli_filter_keywords_by_option(
    keywords: list[KeywordData],
    option: KeywordFilterOption,
    matcher_fn: Callable[[KeywordData], bool],
    descriptor: str,
) -> list[KeywordData]:
    """
    Filter keywords on given condition function. Let the user know what was filtered.
    """
    opt = option.lower()

    if opt == "include":
        return keywords

    if opt == "exclude":
        click.echo(f"{NOTE_MARKER} Excluding {descriptor} keywords")
        return list(filter(lambda kw: matcher_fn(kw) is False, keywords))

    if opt == "only":
        click.echo(f"{NOTE_MARKER} Only showing {descriptor} keywords")
        return list(filter(lambda kw: matcher_fn(kw) is True, keywords))

    msg = f"Unexpected value '{option}' when filtering {descriptor} keywords"
    raise TypeError(msg)



from typing import cast

import click
import robot.errors
from robot.api.parsing import LibraryImport, ModelVisitor
from robot.libdoc import LibraryDocumentation
from robot.libdocpkg.model import KeywordDoc, LibraryDoc

from robotframework_find_unused.common.const import ERROR_MARKER, LibraryData
from robotframework_find_unused.common.convert import libdoc_keyword_to_keyword_data
from robotframework_find_unused.common.enrich_python_keywords import enrich_python_keyword_data
from robotframework_find_unused.common.normalize import normalize_library_name


class LibraryImportVisitor(ModelVisitor):
    """
    Gather downloaded library imports
    """

    downloaded_libraries: dict[str, LibraryData]

    def __init__(self, *, enrich_py_keywords: bool = False) -> None:
        self.enrich_py_keywords = enrich_py_keywords
        self.downloaded_libraries = {}
        super().__init__()

        # Is always imported automatically by Robot
        self._register_downloaded_library("BuiltIn")

    def visit_LibraryImport(self, node: LibraryImport):  # noqa: N802
        """Find out which libraries are actually used"""
        lib_name = node.name

        if lib_name.endswith(".py"):
            # Not a downloaded lib. We already discovered this.
            return

        self._register_downloaded_library(lib_name)

    def _register_downloaded_library(self, lib_name: str) -> None:
        normalized_lib_name = normalize_library_name(lib_name)

        if normalized_lib_name in self.downloaded_libraries:
            # Already found it
            return

        try:
            lib: LibraryDoc = LibraryDocumentation(lib_name)
        except robot.errors.DataError:
            click.echo(
                f"{ERROR_MARKER} Failed to gather keywords from library `{lib_name}`",
                err=True,
            )

            self.downloaded_libraries[normalized_lib_name] = LibraryData(
                name=lib_name,
                name_normalized=normalized_lib_name,
                keywords=[],
                keyword_names_normalized=set(),
            )
            return

        if self.enrich_py_keywords:
            enriched_keywords = enrich_python_keyword_data(lib)
            keywords = [
                libdoc_keyword_to_keyword_data(
                    kw.doc,
                    "LIBRARY",
                    kw.returns,
                )
                for kw in enriched_keywords
            ]
        else:
            lib_keywords = cast(list[KeywordDoc], lib.keywords)
            keywords = [
                libdoc_keyword_to_keyword_data(
                    kw,
                    "LIBRARY",
                )
                for kw in lib_keywords
            ]

        keyword_names_normalized = {kw.normalized_name for kw in keywords}

        self.downloaded_libraries[normalized_lib_name] = LibraryData(
            name=lib_name,
            name_normalized=normalized_lib_name,
            keywords=keywords,
            keyword_names_normalized=keyword_names_normalized,
        )



import ast

from robot.libdocpkg.model import KeywordDoc

from robotframework_find_unused.common.impossible_state_error import ImpossibleStateError


class EnrichedKeywordDoc:
    """Wrap Libdocs KeywordDoc to add more data points."""

    returns: bool | None = None

    def __init__(self, doc: KeywordDoc) -> None:
        self.doc = doc


class PythonKeywordVisitor(ast.NodeVisitor):
    """Visit single Python file AST to find data in functions"""

    def __init__(self, keywords: list[KeywordDoc]) -> None:
        self.keywords: list[EnrichedKeywordDoc] = []
        for keyword in keywords:
            self.keywords.append(EnrichedKeywordDoc(keyword))

    def visit_FunctionDef(self, node: ast.FunctionDef):  # noqa: N802
        """Visit function definition"""
        self._register_function_return(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):  # noqa: N802
        """Visit async function definition"""
        self._register_function_return(node)

    def _register_function_return(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        matching_keywords = [kw for kw in self.keywords if kw.doc.lineno == node.lineno]
        if not matching_keywords:
            # Function is not a keyword
            return

        if len(matching_keywords) > 1:
            msg = "Found multiple Python keyword definitions on the same line"
            raise ImpossibleStateError(msg)

        matching_keyword = matching_keywords[0]

        visitor = PythonKeywordReturnVisitor()
        visitor.visit(node)
        matching_keyword.returns = visitor.has_return_node


class PythonKeywordReturnVisitor(ast.NodeVisitor):
    """Only visit return statements of Python AST."""

    def __init__(self) -> None:
        self.has_return_node: bool = False

    def visit_Return(self, node: ast.Return):  # noqa: N802
        """Visit function return"""
        if self.has_return_node is True:
            return
        self.has_return_node = node.value is not None



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any, Literal

from robot.api.parsing import (
    Keyword,
    KeywordCall,
    ModelVisitor,
    Setup,
    SuiteSetup,
    SuiteTeardown,
    Teardown,
    Template,
    TemplateArguments,
    TestCase,
    TestSetup,
    TestTeardown,
    Token,
)
from robot.libdocpkg.model import ArgumentSpec
from robot.parsing.model.blocks import Block
from robot.running.arguments.argumentmapper import DefaultValue

from robotframework_find_unused.common.const import KeywordData
from robotframework_find_unused.common.normalize import normalize_keyword_name
from robotframework_find_unused.visitors.library_import import LibraryData


@dataclass
class KeywordCallData:
    """
    Simplified keyword call data structure
    """

    keyword: str
    args: tuple[str, ...]


class KeywordVisitor(ModelVisitor):
    """
    A Robot Framework visitor.

    Gathers keywords
    Counts keyword usage
    Counts keyword argument usage
    """

    keywords: dict[str, KeywordData]
    downloaded_libraries: list[LibraryData]
    normalized_keyword_names: set[str]

    def __init__(
        self,
        custom_keywords: list[KeywordData],
        downloaded_library_keywords: list[LibraryData],
    ) -> None:
        self.normalized_keyword_names = set()
        self.keywords = {}
        for kw in custom_keywords:
            self.keywords[kw.normalized_name] = kw
            self.normalized_keyword_names.add(kw.normalized_name)

        self.downloaded_libraries = downloaded_library_keywords
        for lib in self.downloaded_libraries:
            self.normalized_keyword_names.update(lib.keyword_names_normalized)

    def visit_Keyword(self, node: Keyword):  # noqa: N802
        """Keyword definition"""
        keyword = self._get_keyword_data(node.name)
        keyword.returns = self._get_keyword_returns(node)

        return self.generic_visit(node)

    def visit_KeywordCall(self, node: KeywordCall):  # noqa: N802
        """Keyword call / Keyword use"""
        return_assign_token = node.get_token(Token.ASSIGN)

        self._count_keyword_call(
            node.keyword,
            node.args,
            return_value_assigned=(return_assign_token is not None),
        )

        return self.generic_visit(node)

    def visit_Setup(self, node: Setup):  # noqa: N802
        """Count keyword use in test/task/keyword setup"""
        keyword_name_token = node.get_token(Token.NAME)
        if keyword_name_token:
            self._count_keyword_call(str(keyword_name_token), args=[])

        return self.generic_visit(node)

    def visit_Teardown(self, node: Teardown):  # noqa: N802
        """Count keyword use in test/task/keyword teardown"""
        keyword_name_token = node.get_token(Token.NAME)
        if keyword_name_token:
            self._count_keyword_call(str(keyword_name_token), args=[])

        return self.generic_visit(node)

    def visit_TestSetup(self, node: TestSetup):  # noqa: N802
        """Count keyword use in test setup"""
        self._count_keyword_call(node.name, node.args)

        return self.generic_visit(node)

    def visit_SuiteSetup(self, node: SuiteSetup):  # noqa: N802
        """Count keyword use in suite setup"""
        self._count_keyword_call(node.name, node.args)

        return self.generic_visit(node)

    def visit_TestTeardown(self, node: TestTeardown):  # noqa: N802
        """Count keyword use in test teardown"""
        self._count_keyword_call(node.name, node.args)

        return self.generic_visit(node)

    def visit_SuiteTeardown(self, node: SuiteTeardown):  # noqa: N802
        """Count keyword use in suite teardown"""
        self._count_keyword_call(node.name, node.args)

    def visit_TestCase(self, node: TestCase):  # noqa: N802
        """Count templated test cases"""
        template_keyword = None
        template_args_set = []
        for child in node.body:
            if isinstance(child, Template):
                template_keyword = str(child.get_token(Token.NAME))
                continue
            if isinstance(child, TemplateArguments):
                args = child.get_tokens(Token.ARGUMENT)
                args = [str(arg) for arg in args]
                template_args_set.append(args)
                continue

        if template_keyword is not None:
            self._count_keyword_call(template_keyword, ())
            for args in template_args_set:
                self._count_keyword_call(template_keyword, args, count_keyword=False)

        return self.generic_visit(node)

    def _remove_lib_from_name(self, name: str) -> str:
        if "." in name:
            name = name.split(".", 1)[1]
        return name

    def _get_keyword_data(self, name: str) -> KeywordData:
        name = self._remove_lib_from_name(name)
        normalized_name = normalize_keyword_name(name)

        if normalized_name not in self.keywords:
            # Found a previously unused:
            # - downloaded library keyword
            # - out-of-scope keyword
            # or found a:
            # - keyword with embedded argument

            if normalized_name in self.normalized_keyword_names:
                self._register_downloaded_library_keyword(name, normalized_name)
            else:
                self._register_unknown_keyword(name, normalized_name)

        return self.keywords[normalized_name]

    def _count_keyword_call(
        self,
        name: str,
        args: Iterable[str],
        *,
        return_value_assigned: bool = False,
        count_keyword: bool = True,
    ) -> None:
        """
        Count the keyword.

        For keywords that take other keywords as arguments: Recursively handle inner keyword.
        """
        keyword = self._get_keyword_data(name)
        if count_keyword:
            keyword.use_count += 1

        if return_value_assigned:
            keyword.return_use_count += 1

        inner_keywords = self._get_keyword_reference_in_argument(args, keyword)
        for inner in inner_keywords:
            self._count_keyword_call(inner.keyword, inner.args)

        self._count_keyword_call_args(keyword, args)

    def _get_keyword_reference_in_argument(
        self,
        args: Iterable[str],
        keyword: KeywordData,
    ) -> list[KeywordCallData]:
        """
        Return keyword calls in the given arguments

        - Only considers known keywords
        - Only considers cases where the keyword name or argument name includes 'keyword'

        Returns a list of tuples where (inner_keywor_name, inner_keyword_arguments)
        """
        inner_keywords: list[KeywordCallData] = []

        args = tuple(a for a in args)
        for i, arg in enumerate(args):
            arg_value = self._argument_is_keyword_reference(arg, i, keyword)
            if arg_value is False:
                continue

            inner_keywords.append(KeywordCallData(keyword=arg_value, args=args[i + 1 :]))

        if len(inner_keywords) > 1:
            for i in range(1, len(inner_keywords)):
                cur = inner_keywords[i]
                prev = inner_keywords[i - 1]

                prev.args = self._get_deduped_arguments(
                    prev,
                    cur,
                )

        return inner_keywords

    def _argument_is_keyword_reference(
        self,
        arg: str,
        position_index: int,
        keyword: KeywordData,
    ) -> Literal[False] | str:
        arg_name = None
        arg_val = arg
        if "=" in arg:
            # Is a named arg
            (arg_name, arg_val) = arg.split("=", 1)

        if "." in arg_val:
            # Remove library prefix
            arg_val = arg_val.split(".", 1)[1]

        normalized_name = normalize_keyword_name(arg_val)
        if normalized_name not in self.normalized_keyword_names:
            # Not a known keyword name
            return False

        if "keyword" in keyword.normalized_name:
            return arg_val

        if arg_name is None:
            # Argument is positional. Named arg can't get here.
            arg_name = self._get_keyword_arg_name_by_position_index(
                keyword,
                position_index=position_index,
            )
        if arg_name is None:
            return False

        arg_name = arg_name.lower()
        if "keyword" in arg_name:
            return arg_val

        return False

    def _get_keyword_arg_name_by_position_index(
        self,
        keyword: KeywordData,
        position_index: int,
    ) -> str | None:
        """
        Return the argument name at the given positional index.

        Does not consider named arguments.
        """
        if keyword.arguments is None:
            # We don't know anything about the defined keyword arguments
            return None

        if keyword.arguments.var_positional and position_index > len(
            keyword.arguments.positional,
        ):
            # @{varargs} used
            return keyword.arguments.var_positional

        if position_index < len(keyword.arguments.argument_names):
            return keyword.arguments.argument_names[position_index]

        return None

    def _get_deduped_arguments(
        self,
        call: KeywordCallData,
        duplicate_call: KeywordCallData,
    ) -> tuple[str, ...]:
        """
        Deduplicate arguments of the first keyword call.

        Used to prevent counting a keyword multiple times.
        """
        args_to_remove = (duplicate_call.keyword, *duplicate_call.args)

        output = [*call.args]
        for i in range(len(args_to_remove)):
            remove_val = args_to_remove[-i - 1]
            if "=" in remove_val:
                (_, remove_val) = remove_val.split("=", 1)

            actual_val = output.pop()
            if "=" in actual_val:
                (_, actual_val) = actual_val.split("=", 1)

            if remove_val == actual_val:
                continue
            if "." in remove_val and remove_val.endswith("." + actual_val):
                continue
            if "." in actual_val and actual_val.endswith("." + remove_val):
                continue

            msg = f"Expected list to end with '{remove_val}', but found '{actual_val}' instead"
            raise ValueError(msg)

        return tuple(output)

    def _register_downloaded_library_keyword(self, name: str, normalized_name: str) -> None:
        """
        Register as a downloaded library keyword.
        """
        library = None
        for lib in self.downloaded_libraries:
            if normalized_name in lib.keyword_names_normalized:
                library = lib
                break

        if library is None:
            msg = f"Can't find library for keyword '{name}'"
            raise Exception(msg)  # noqa: TRY002

        library_keyword = None
        for kw in library.keywords:
            if kw.normalized_name == normalized_name:
                library_keyword = kw
                break

        if library_keyword is None:
            msg = f"Can't find keyword '{name}' in library '{library.name}'"
            raise Exception(msg)  # noqa: TRY002

        self.keywords[library_keyword.normalized_name] = library_keyword

    def _register_unknown_keyword(self, name: str, normalized_name: str) -> None:
        """
        Register as an unknown keyword with minimum data that does not look weird.
        """
        self.keywords[normalized_name] = KeywordData(
            name=name,
            normalized_name=normalized_name,
            argument_use_count=None,
            deprecated=None,
            private=False,
            use_count=0,
            returns=None,
            return_use_count=0,
            type="UNKNOWN",
            arguments=None,
            library="",
        )

    def _parse_args(
        self,
        call_args: Iterable[str],
        kw_args: ArgumentSpec,
    ) -> tuple[list[str], list[tuple[str, Any]]]:
        positional_args: list[str] = []
        named_args: list[tuple[str, Any]] = []

        for arg in call_args:
            if "=" not in arg:
                positional_args.append(arg)
                continue

            (named_arg_name, named_arg_val) = arg.split("=", 1)
            if named_arg_name in kw_args.argument_names:
                # It's a correct named argument
                named_args.append((named_arg_name, named_arg_val))
            else:
                positional_args.append(arg)

        return (positional_args, named_args)

    def _count_keyword_call_args(self, kw: KeywordData, call_args: Iterable[str]) -> None:
        if not kw.arguments or kw.argument_use_count is None:
            # This is a downloaded library keyword. We don't care about the args
            return
        (positional_args, named_args) = self._parse_args(call_args, kw.arguments)

        (called_with_args, called_with_kwargs) = kw.arguments.map(
            positional_args,
            named_args,
            replace_defaults=False,
        )

        called_with_kwarg_names = [a[0] for a in called_with_kwargs]
        kw_arg_names = [a for a in kw.arguments.argument_names if a not in called_with_kwarg_names]

        if len(kw_arg_names) == 0:
            return

        for position, arg in enumerate(called_with_args):
            if isinstance(arg, DefaultValue):
                continue

            if position >= len(kw_arg_names):
                position = len(kw_arg_names) - 1  # noqa: PLW2901

            arg_name = kw_arg_names[position]
            kw.argument_use_count[arg_name] += 1

        for name, val in called_with_kwargs:
            if isinstance(val, DefaultValue):
                continue
            kw.argument_use_count[name] += 1

    def _get_keyword_returns(self, node: Keyword | Block) -> bool:  # noqa: C901
        """
        Return if keyword returns a value or not.

        A return must have an explicit value.
        """
        for token in node.body:
            if isinstance(token, Block):
                # Block like `IF`, `FOR`, etc. Crawl recursively
                block_returns = self._get_keyword_returns(token)
                if block_returns is True:
                    return True
                continue

            if token.type in (
                Token.RETURN,
                Token.RETURN_STATEMENT,
            ):
                # `RETURN` and `[Return]` syntax
                if token.get_token(Token.ARGUMENT) is not None:
                    return True
                continue

            if token.type == Token.KEYWORD:
                called_keyword_name = token.get_token(Token.KEYWORD)
                if not called_keyword_name or called_keyword_name.value is None:
                    continue

                # Node is special return keywords `Return From Keyword` and `Return From Keyword If`

                keyword_name_normalized = normalize_keyword_name(called_keyword_name.value)
                if keyword_name_normalized not in (
                    "returnfromkeyword",
                    "returnfromkeywordif",
                ):
                    continue

                argument_count = len(token.get_tokens(Token.ARGUMENT))
                if keyword_name_normalized == "returnfromkeyword" and argument_count >= 1:
                    return True
                if keyword_name_normalized == "returnfromkeywordif" and argument_count >= 2:  # noqa: PLR2004
                    return True
                continue

        return False



def normalize_variable_name(name: str, *, strip_decoration: bool = True) -> str:
    """
    Normalize Robot variables name. Output is suitable for matching and sorting purposes
    """
    norm = name.replace(" ", "").replace("_", "").lower()

    if strip_decoration and norm[0] in ("$", "@", "&", "%"):
        norm = norm[1:].removeprefix("{").removesuffix("}")

    return norm


def normalize_keyword_name(name: str) -> str:
    """
    Normalize Robot keyword name. Output is suitable for matching and sorting purposes
    """
    return name.replace(" ", "").replace("_", "").lower()


def normalize_library_name(name: str) -> str:
    """
    Normalize Robot library name. Output is suitable for matching and sorting purposes
    """
    return normalize_keyword_name(name)



from pathlib import Path

import click

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    VERBOSE_NO,
    VERBOSE_SINGLE,
    VariableData,
)
from robotframework_find_unused.common.normalize import normalize_variable_name
from robotframework_find_unused.common.parse import resolve_variable_name
from robotframework_find_unused.common.visit import visit_robot_files
from robotframework_find_unused.visitors.variable_definition import VariableDefinitionVisitor


def cli_get_variable_definitions(
    file_paths: list[Path],
    *,
    verbose: int,
):
    """
    Walk through all robot files to discover non-local variable definitions and show progress
    """
    click.echo("Gathering variables definitions...")
    variables = _get_variable_definitions(file_paths)

    _log_variable_stats(list(variables.values()), verbose)
    return variables


def _get_variable_definitions(file_paths: list[Path]) -> dict[str, VariableData]:
    """
    Walk through all robot files to discover non-local variable definitions.
    """
    visitor = VariableDefinitionVisitor()
    visit_robot_files(file_paths, visitor)

    return _resolve_vars_in_var_name(visitor.variables)


def _resolve_vars_in_var_name(variables: dict[str, VariableData]) -> dict[str, VariableData]:
    resolved_variables: dict[str, VariableData] = {}
    all_used_vars: list[str] = []
    for var in variables.values():
        var_name = var.normalized_name
        (resolved_var_name, used_vars) = resolve_variable_name(var_name, variables)
        resolved_var_name_normalized = normalize_variable_name(resolved_var_name)

        if resolved_var_name_normalized == var_name:
            # Nothing to resolve
            resolved_variables[var_name] = var
            continue

        resolved_variables[resolved_var_name_normalized] = VariableData(
            name=var.name,
            normalized_name=resolved_var_name_normalized,
            resolved_name="${" + resolved_var_name + "}",
            use_count=var.use_count,
            defined_in_type=var.defined_in_type,
            defined_in=var.defined_in,
            value=var.value,
        )

        all_used_vars = all_used_vars + used_vars

    for used_var in all_used_vars:
        if used_var not in resolved_variables:
            continue
        resolved_variables[used_var].use_count += 1

    return resolved_variables


def _log_variable_stats(variables: list[VariableData], verbose: int) -> None:
    """
    Output details to the user
    """
    click.echo(
        (ERROR_MARKER if len(variables) == 0 else DONE_MARKER)
        + f" Found {len(variables)} unique non-local variables definitions",
    )

    if verbose == VERBOSE_NO:
        return

    var_types: dict[str, list[str]] = {}
    for var in variables:
        if var.defined_in_type not in var_types:
            var_types[var.defined_in_type] = []
        var_types[var.defined_in_type].append(var.name)

    for defined_in_type, var_names in sorted(
        var_types.items(),
        key=lambda items: len(items[1]),
        reverse=True,
    ):
        click.echo(f"{INDENT}{len(var_names)} variables definitions of type '{defined_in_type}'")

        if verbose == VERBOSE_SINGLE:
            continue
        for name in var_names:
            click.echo(f"{INDENT}{INDENT}{click.style(name, fg='bright_black')}")



from collections.abc import Iterable
from pathlib import Path

import click
import robot.errors
from robot.api.parsing import (
    File,
    KeywordSection,
    ModelVisitor,
    TestCaseSection,
    Variable,
    VariableSection,
    VariablesImport,
)

from robotframework_find_unused.common.const import (
    ERROR_MARKER,
    VariableData,
    VariableDefinedInType,
)
from robotframework_find_unused.common.impossible_state_error import ImpossibleStateError
from robotframework_find_unused.common.normalize import normalize_variable_name


class VariableDefinitionVisitor(ModelVisitor):
    """
    Visit file and discover variable definitions.
    """

    variables: dict[str, VariableData]
    current_working_file: Path | None = None
    current_working_directory: Path | None = None

    def __init__(self) -> None:
        self.variables = {}
        super().__init__()

    def visit_File(self, node: File):  # noqa: N802
        """Keep track of the current working file"""
        if node.source is not None:
            self.current_working_file = node.source
            self.current_working_directory = self.current_working_file.parent

        return self.generic_visit(node)

    def visit_VariableSection(self, node: VariableSection):  # noqa: N802
        """
        Look for variable declarations in the variables section.
        """
        if self.current_working_file is None:
            msg = "Found variables section outside a .robot or .resource file"
            raise ImpossibleStateError(msg)

        for var_node in node.body:
            if not isinstance(var_node, Variable):
                continue

            self._register_variable(
                var_node.name,
                "variables_section",
                self.current_working_file,
                var_node.value,
            )

        return self.generic_visit(node)

    def visit_VariablesImport(self, node: VariablesImport):  # noqa: N802
        """
        Look for variable declarations in variable files.
        """
        if self.current_working_directory is None:
            msg = "Found variables file import outside a .robot or .resource file"
            raise ImpossibleStateError(msg)

        import_path = node.name
        if "/" in import_path or "\\" in import_path:
            # Is a file path. Make it absolute
            import_path = self.current_working_directory.joinpath(node.name)

        try:
            self._import_variable_file(Path(import_path), node.args)
        except Exception as e:  # noqa: BLE001
            click.echo(f"{ERROR_MARKER} Failed to import variables from variables file.")
            click.echo(f"{ERROR_MARKER} Something went very wrong. Details below:")
            click.echo(f"{ERROR_MARKER} {e}")
            click.echo()

        return self.generic_visit(node)

    def visit_KeywordSection(self, _: KeywordSection):  # noqa: N802
        """Don't visit anything inside keyword sections. We don't need it"""
        return

    def visit_TestCaseSection(self, _: TestCaseSection):  # noqa: N802
        """Don't visit anything inside test case sections. We don't need it"""
        return

    def _import_variable_file(self, import_path: Path, import_args: tuple[str, ...]) -> None:
        """
        Import a file as a variable file.

        WARNING: This function uses code that is NOT in the public Robot API.
        Always wrap this function in a try-except to reduce the impact of internal Robot code
        changing.
        """
        from robot.variables.filesetter import VariableFileSetter
        from robot.variables.store import VariableStore

        var_store = VariableStore(None)
        file_setter = VariableFileSetter(var_store)

        try:
            file_setter.set(
                str(import_path),
                args=import_args,
            )
        except robot.errors.DataError as e:
            click.echo(f"{ERROR_MARKER} {e.message.splitlines()[0]}")
            return

        for var_name in var_store.as_dict(decoration=True):
            self._register_variable(var_name, "variable_file", import_path, [])

    def _register_variable(
        self,
        name: str,
        defined_in_type: VariableDefinedInType,
        defined_in: Path,
        value: Iterable[str],
    ) -> None:
        name_normalized = normalize_variable_name(name)
        if name_normalized in self.variables:
            var_def = self.variables[name_normalized]

            if var_def.defined_in_type in (
                "variables_section",
                "variable_file",
            ):
                # Existing def is from primary source. Keep
                return

            if var_def.defined_in_type == defined_in_type:
                # Existing def is from equal source. Keep
                return

        self.variables[name_normalized] = VariableData(
            name=name,
            normalized_name=name_normalized,
            resolved_name=name,
            use_count=0,
            defined_in_type=defined_in_type,
            defined_in=defined_in.as_posix(),
            value=value,
        )



ISSUE_TRACKER_URL = "https://github.com/Lakitna/robotframework-find-unused/issues"


class ImpossibleStateError(Exception):
    """
    Exception for things that should not be possible.

    This error being raised means that something went horribly wrong somewhere.
    """

    def __init__(self, message: str) -> None:
        message = (
            message
            + " - This should be impossible. Please open an issue at "
            + ISSUE_TRACKER_URL
            + " and let us know how you managed to do this."
        )
        super().__init__(message)



import re
from collections.abc import Iterable

from robot.api.parsing import (
    Arguments,
    For,
    If,
    KeywordCall,
    ModelVisitor,
    Variable,
    VariableSection,
)

from robotframework_find_unused.common.const import VariableData
from robotframework_find_unused.common.normalize import normalize_variable_name
from robotframework_find_unused.common.parse import (
    get_variables_in_string,
    resolve_variable_name,
    supported_builtin_vars,
)


class VariableCountVisitor(ModelVisitor):
    """
    Visit file and count variable usage.
    """

    variables: dict[str, VariableData]

    # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#special-variable-syntax
    _pattern_eval_variable = re.compile(r"\$(\w+)")
    # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#inline-python-evaluation
    _pattern_inline_eval = re.compile(r"\${{(.+?)}}")

    def __init__(self, variable_defs: dict[str, VariableData]) -> None:
        self.variables = variable_defs
        super().__init__()

    def visit_VariableSection(self, node: VariableSection):  # noqa: N802
        """
        Look for used variables in variable definitions.
        """
        for var_node in node.body:
            if not isinstance(var_node, Variable):
                continue
            self._count_used_vars_in_args(var_node.value)

        return self.generic_visit(node)

    def visit_Arguments(self, node: Arguments):  # noqa: N802
        """
        Look for used variables in the default value of keyword arguments.
        """
        arguments = node.values

        for argument in arguments:
            if "=" not in argument:
                # Argument has no default. We don't care about it.
                continue

            argument_default = argument.split("=", 1)[1]
            self._count_used_vars_in_args([argument_default])

        return self.generic_visit(node)

    def visit_KeywordCall(self, node: KeywordCall):  # noqa: N802
        """
        Look for used variables called keyword arguments.
        """
        if node.keyword.lower() == "evaluate":
            self._count_used_vars_in_eval(node.args[0])
        else:
            self._count_used_vars_in_args(node.args)

        return self.generic_visit(node)

    def visit_For(self, node: For):  # pyright: ignore[reportIncompatibleMethodOverride] # noqa: N802
        """
        Look for used variables in for loop conditions.
        """
        self._count_used_vars_in_args(node.values)

        return self.generic_visit(node)

    def visit_If(self, node: If):  # pyright: ignore[reportIncompatibleMethodOverride] # noqa: N802
        """
        Look for used variables in if/else/elseif statement conditions.
        """
        if node.condition:
            self._count_used_vars_in_eval(node.condition)

        return self.generic_visit(node)

    def _count_used_vars_in_eval(self, eval_str: str) -> None:
        """
        Count used variables found in a python evaluation context
        """
        used_vars = self._get_used_vars_in_eval(eval_str)
        used_vars = self._filter_supported_vars(used_vars)
        for name in used_vars:
            self._count_variable_use(name)

    def _get_used_vars_in_eval(self, eval_str: str) -> list[str]:
        """
        Return a list of used variables in a given evaluated Python expression
        """
        eval_str = eval_str.strip()
        used_vars = self._get_used_vars_in_args([eval_str])

        match = self._pattern_eval_variable.findall(eval_str)
        for var in match:
            used_vars.append("${" + normalize_variable_name(var) + "}")

        return used_vars

    def _count_used_vars_in_args(self, args: Iterable[str]) -> None:
        """
        Count used variables found in a list of arguments
        """
        used_vars = self._get_used_vars_in_args(args)
        used_vars = self._filter_supported_vars(used_vars)
        for name in used_vars:
            self._count_variable_use(name)

    def _get_used_vars_in_args(self, args: Iterable[str]) -> list[str]:
        """
        Return a list of used variables in a given list of strings
        """
        used_vars = []
        for arg in args:
            var_match = get_variables_in_string(arg)
            used_vars += var_match

            eval_match = self._pattern_inline_eval.findall(arg)
            for inline_eval in eval_match:
                used_vars += self._get_used_vars_in_eval(inline_eval)

        return used_vars

    def _filter_supported_vars(self, variables: list[str]) -> list[str]:
        """
        Filter out unsupported variables and some Robot builtin stuff.
        """
        filtered = []
        for formatted_var in variables:
            var = normalize_variable_name(formatted_var)

            try:
                float(var)
                # Is a number, not a variable name.
                # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#number-variables
                continue
            except ValueError:
                pass

            if var in supported_builtin_vars:
                continue

            (var, used_vars) = resolve_variable_name(var, self.variables)
            for v in used_vars:
                self._count_variable_use(v)

            if not var.isalnum():
                # Potential extended variable syntax
                var = self._normalize_extended_variable_syntax(var)

            filtered.append(var)

        return filtered

    def _normalize_extended_variable_syntax(self, var: str) -> str:
        if var in self.variables:
            return var

        var_name = var
        while len(var_name) > 0:
            # Remove all trailing alphanumeric
            while len(var_name) > 0 and var_name[-1].isalnum():
                var_name = var_name[0:-1]
            if len(var_name) == 0:
                break

            # Remove single trailing special char
            var_name = var_name[0:-1]
            if len(var_name) == 0:
                break

            if var_name in self.variables:
                return var_name

        # Could not find var. Don't modify.
        return var

    def _count_variable_use(self, normalized_name: str) -> None:
        """
        Count the variable.
        """
        if normalized_name not in self.variables:
            # Unknown variable definition. Ignore
            return
        self.variables[normalized_name].use_count += 1



import ast
from pathlib import Path
from typing import cast

from robot.libdocpkg.model import LibraryDoc

from robotframework_find_unused.visitors.python_keyword_visitor import (
    EnrichedKeywordDoc,
    PythonKeywordVisitor,
)


def enrich_python_keyword_data(libdoc: LibraryDoc) -> list[EnrichedKeywordDoc]:
    """Gather data on Python keyword returns"""
    source_path = Path(cast(str, libdoc.source))
    with source_path.open() as f:
        raw_python_source = f.read()
    model = ast.parse(raw_python_source)

    visitor = PythonKeywordVisitor(libdoc.keywords)
    visitor.visit(model)

    return visitor.keywords



"""
CLI entry point
"""

# ruff: noqa: FBT001,D301

import sys

import click

from .commands import (
    ArgumentsOptions,
    KeywordOptions,
    ReturnOptions,
    VariableOptions,
    cli_arguments,
    cli_keywords,
    cli_returns,
    cli_variables,
)
from .common.const import KeywordFilterOption

click_choice_keyword_filter_option = click.Choice(
    ["include", "exclude", "only"],
    case_sensitive=False,
)


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
def cli():
    """
    Find unused parts of your Robot Framework project.
    """


@cli.command(name="keywords")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Output usage count for all keywords instead of only unused keywords",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help="Only output keywords who's name match the glob pattern. Match without library prefix",
)
@click.option(
    "-d",
    "--deprecated",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-l",
    "--library",
    type=click_choice_keyword_filter_option,
    default="exclude",
    show_default=True,
    help="How to output keywords from downloaded libraries",
)
@click.option(
    "-u",
    "--unused-library",
    type=click.Choice(["include", "exclude"], case_sensitive=False),
    default="exclude",
    show_default=True,
    help="How to output unused keywords from downloaded libraries",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    count=True,
    help="Show more log output. Can be used twice",
)
@click.argument("file_path", default=".")
def keywords(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: KeywordFilterOption,
    private: KeywordFilterOption,
    library: KeywordFilterOption,
    unused_library: KeywordFilterOption,
    verbose: int,
    file_path: str,
):
    """
    Find unused keywords

    Traverse files in the given file path. In those files, count how often each keyword is used.
    Keywords with 0 uses are logged.

    ----------

    Limitation 1: Keywords with embedded arguments are not counted

    Example: This keyword is not counted because it contains the embedded argument ${something}:

    \b
        Do ${something} amazing

    ----------

    Limitation 2: Library prefixes are ignored.

    Example: The following keywords are counted as the same keyword:

    \b
        SeleniumLibrary.Get Text
        AppiumLibrary.Get Text

    ----------

    Limitation 3: Most keywords used as an argument for another keyword are counted, but some may
    not be.

    Example: 'Beautiful keyword' is not counted.

    \b
        Do Something Amazing    ${True}    Beautiful keyword

    To ensure that your keyword in an argument is counted, your keyword name or argument name
    must include the literal word 'keyword' (case insensitive).

    Example: 'Beautiful keyword' is counted, because 'Run Keyword' includes the word 'keyword'

    \b
        Run Keyword    Beautiful keyword

    Example: 'Beautiful keyword' is counted, because the argument ${inner_keyword} includes the word
    'keyword'

    \b
        Amazing    ${True}    inner_keyword=Beautiful keyword
    """
    options = KeywordOptions(
        source_path=file_path,
        deprecated_keywords=deprecated,
        private_keywords=private,
        library_keywords=library,
        unused_library_keywords=unused_library,
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    exit_code = cli_keywords(options)
    sys.exit(exit_code)


@cli.command(name="variables")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Show usage count for all variables instead of only unused variables",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help=(
        "Only show variables who's name match the glob pattern. "
        "Matching without {brackets} and $@&% prefixes"
    ),
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    count=True,
    help="Show more log output. Can be used twice",
)
@click.argument("file_path", default=".")
def variables(
    show_count: bool,
    filter: str | None,  # noqa: A002
    verbose: int,
    file_path: str,
):
    """
    Find unused global variables

    Traverse files in the given file path. In those files, count how often each global variable is
    used. Variables defined in a variables section or variable file with 0 uses are logged.

    ----------

    Limitation 1: Only globally user-defined variables.

    All of the following variables are ignored:

    \b
    - Variables only provided via the command line
    - Environment variables
    - BuiltIn variables
    - Variables only set with `Set Global Variable`
    - Variables only set with `VAR  ...  scope=GLOBAL`
    - Variables only set with `Set Suite Variable`
    - Variables only set with `VAR  ...  scope=SUITE`
    - Variables only set with `Set Test Variable`
    - Variables only set with `VAR  ...  scope=TEST`
    - Variables only set with `Set Task Variable`
    - Variables only set with `VAR  ...  scope=TASK`
    - Variables only set with `Set Variable`
    - Variables only set with `VAR  ...`
    - Variables only set with the return value of a keyword

    ----------

    Limitation 2: Variables with variables in their name are not always counted.

    When using or defining a variable, the variable name can contain other variables. The most
    common usecases are supported but there are a lot of possible complexity which is not supported.

    Variables in variable names are only counted when:

    \b
    - None of the involved variables are limited by limitation 1.
    - All nested variables are single-line scalar variables.
      - No lists (e.g. `@{example}`)
      - No dicts (e.g. `&{example}`)
      - No multi-line string definitions
    - No extended variable syntax.

    Example: ${hello_${place.lower()}} uses extended variable syntax and is therefore ignored.
    Because of this, the variable ${hello_world} will be falsely flagged as unused.

    \b
        *** Variables ***
        ${hello_world}    Hello World!
        ${place}          WORLD

    \b
        *** Keywords ***
        My Amazing Keyword
            Log    ${hello_${place.lower()}}

    ----------

    Limitation 3: Only counts variable uses in `.robot` and `.resource` files.

    Using variables in Python files is never counted. This is true for both libraries and Python
    variable files.

    Example: The use of the variable `person` is not counted because it's used in a Python variable
    file.

    \b
        person = "Pekka"
        message = "Hello " + person
    """
    options = VariableOptions(
        source_path=file_path,
        show_all_count=show_count,
        filter_glob=filter,
        verbose=verbose,
    )
    exit_code = cli_variables(options)
    sys.exit(exit_code)


@cli.command(name="arguments")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Show usage count for all arguments instead of only unused arguments",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help="Only output keywords who's name match the glob pattern. Match without library prefix",
)
@click.option(
    "-d",
    "--deprecated",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-l",
    "--library",
    type=click_choice_keyword_filter_option,
    default="exclude",
    show_default=True,
    help="How to output keywords from downloaded libraries",
)
@click.option(
    "-u",
    "--unused",
    type=click_choice_keyword_filter_option,
    default="exclude",
    show_default=True,
    help="How to output unused keywords",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    count=True,
    help="Show more log output. Can be used twice",
)
@click.argument("file_path", default=".")
def arguments(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: KeywordFilterOption,
    private: KeywordFilterOption,
    library: KeywordFilterOption,
    unused: KeywordFilterOption,
    verbose: int,
    file_path: str,
):
    """
    Find unchanged default keyword arguments

    Traverse files in the given file path. In those files, count how often each argument is used
    during a keyword call. Arguments with 0 uses are logged.

    ----------

    Limitation 1: Arguments for keywords with embedded arguments are not counted

    Example: The argument ${beautiful} is not counted because the keyword contains the embedded
    argument ${something}:

    \b
        Do ${something} amazing
            [Arguments]    ${beautiful}=${True}

    ----------

    Limitation 2: Most keywords used as an argument for another keyword are counted, but some may
    not be. This includes the arguments used by the inner keyword.

    Example: 'Beautiful keyword' is not recognized as a keyword. Because of this, the ${hello}
    argument of 'Beautiful keyword' is falsely counted as an argument for 'Do Something Amazing'.

    \b
        Do Something Amazing    Beautiful keyword    hello=${True}

    To ensure that your keyword is handled properly, your keyword name or argument name must include
    the literal word 'keyword' (case insensitive).

    Example: The ${hello} argument of 'Beautiful keyword' is counted, because 'Run Keyword' includes
    the word 'keyword'

    \b
        Run Keyword    Beautiful keyword    hello=${True}

    Example: The ${hello} argument of 'Beautiful keyword' is counted, because the argument
    ${inner_keyword} includes the word 'keyword'.

    \b
        Amazing    inner_keyword=Beautiful keyword    hello=${True}

    Note how the script assumes that all arguments after ${inner_keyword} are arguments for
    'Beautiful keyword'.
    """
    options = ArgumentsOptions(
        source_path=file_path,
        deprecated_keywords=deprecated,
        private_keywords=private,
        library_keywords=library,
        unused_keywords=unused,
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    exit_code = cli_arguments(options)
    sys.exit(exit_code)


@cli.command(name="returns")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Output usage count for all keywords instead of only keywords with unused returns",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help="Only output keywords who's name match the glob pattern. Match without library prefix",
)
@click.option(
    "-d",
    "--deprecated",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click_choice_keyword_filter_option,
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-l",
    "--library",
    type=click_choice_keyword_filter_option,
    default="exclude",
    show_default=True,
    help="How to output keywords from downloaded libraries",
)
@click.option(
    "-u",
    "--unused",
    type=click_choice_keyword_filter_option,
    default="exclude",
    show_default=True,
    help="How to output unused keywords",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    count=True,
    help="Show more log output. Can be used twice",
)
@click.argument("file_path", default=".")
def returns(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: KeywordFilterOption,
    private: KeywordFilterOption,
    library: KeywordFilterOption,
    unused: KeywordFilterOption,
    verbose: int,
    file_path: str,
):
    """
    Find unused keyword return values

    Traverse files in the given file path. In those files, count how often each keyword return
    value is used. Keywords whose return value is never used are logged.

    ----------

    Limitation 1: Return value not counted when the keyword is used as an argument for another
    keyword.

    Example: The return value of 'Beautiful keyword' is not counted.

    \b
        ${returned_value} =    Run Keyword    Beautiful keyword

    This situation can't be counted without knowing what exactly `Run Keyword` does.
    """
    options = ReturnOptions(
        source_path=file_path,
        deprecated_keywords=deprecated,
        private_keywords=private,
        library_keywords=library,
        unused_keywords=unused,
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    exit_code = cli_returns(options)
    sys.exit(exit_code)



import sys
from typing import Literal

import click

from .const import NOTE_MARKER, VERBOSE_DOUBLE, KeywordData


def pretty_kw_name(keyword: KeywordData) -> str:
    """
    Format keyword name for output to the user
    """
    name = keyword.name

    if keyword.library:
        name = click.style(keyword.library + ".", fg="bright_black") + name

    if keyword.deprecated is True:
        name += " " + click.style("[DEPRECATED]", fg="red")

    return name


def cli_hard_exit(verbose: int) -> Literal[255]:
    """
    Immediately hard exit app. Use when something went wrong.
    """
    if verbose < VERBOSE_DOUBLE:
        click.echo(f"{NOTE_MARKER} Run with `--verbose --verbose` or `-vv` for more details")
    sys.exit(255)
    return 255



from typing import Any, Literal, cast

from robot.libdocpkg.model import KeywordDoc

from .const import KeywordData
from .normalize import normalize_keyword_name


def libdoc_keyword_to_keyword_data(
    libdoc: KeywordDoc,
    keyword_type: Literal["CUSTOM_SUITE", "CUSTOM_LIBRARY", "CUSTOM_RESOURCE", "LIBRARY"],
    keyword_returns: bool | None = None,
):
    """
    Convert a Libdoc keyword to the internally used data structure
    """
    argument_use_count = {}
    for arg in libdoc.args.argument_names:
        argument_use_count[arg] = 0

    return KeywordData(
        normalized_name=normalize_keyword_name(libdoc.name),
        name=libdoc.name,
        library=cast(Any, libdoc.parent).name,
        deprecated=(libdoc.deprecated is True),
        private=("robot:private" in libdoc.tags),
        argument_use_count=argument_use_count,
        arguments=libdoc.args,
        use_count=0,
        returns=keyword_returns,
        return_use_count=0,
        type=keyword_type,
    )



"""
Implementation of the 'variables' command
"""

import fnmatch
from dataclasses import dataclass

import click

from robotframework_find_unused.common.cli import cli_hard_exit
from robotframework_find_unused.common.const import INDENT, VariableData

from .step.discover_files import cli_discover_file_paths
from .step.variables_count_uses import cli_count_variable_uses
from .step.variables_definitions import cli_get_variable_definitions


@dataclass
class VariableOptions:
    """
    Command line options for the 'variables' command
    """

    show_all_count: bool
    filter_glob: str | None
    verbose: int
    source_path: str


def cli_variables(options: VariableOptions):
    """
    Entry point for the CLI command
    """
    file_paths = cli_discover_file_paths(options.source_path, verbose=options.verbose)
    if len(file_paths) == 0:
        return cli_hard_exit(options.verbose)

    variables = cli_get_variable_definitions(file_paths, verbose=options.verbose)
    if len(variables) == 0:
        return cli_hard_exit(options.verbose)

    variables = cli_count_variable_uses(
        file_paths,
        variables,
        verbose=options.verbose,
    )

    _cli_log_results(variables, options)
    return _exit_code(variables)


def _cli_log_results(variables: list[VariableData], options: VariableOptions) -> None:
    click.echo()

    if options.filter_glob:
        click.echo(f"Only showing variables matching pattern '{options.filter_glob}'")

        pattern = options.filter_glob.lower()
        filtered_variables = []
        for var in variables:
            if fnmatch.fnmatchcase(var.normalized_name, pattern):
                filtered_variables.append(var)

        variables = filtered_variables

    if options.show_all_count:
        sorted_variables = sorted(variables, key=lambda var: var.normalized_name)
        sorted_variables = sorted(sorted_variables, key=lambda var: var.use_count)

        click.echo("use_count\tvariable")
        for var in sorted_variables:
            name = var.name
            if var.name != var.resolved_name:
                name += click.style(f" -> {var.resolved_name}", fg="bright_black")
            click.echo("\t".join([str(var.use_count), name]))
    else:
        unused_variables = [var for var in variables if var.use_count == 0]
        unused_variables = sorted(unused_variables, key=lambda var: var.normalized_name)

        click.echo(f"Found {len(unused_variables)} unused variables:")
        for var in unused_variables:
            click.echo(INDENT + var.name)


def _exit_code(variables: list[VariableData]) -> int:
    unused_variables = [var for var in variables if var.use_count == 0]
    exit_code = len(unused_variables)
    return min(exit_code, 200)



"""
Implementation of the 'returns' command
"""

from dataclasses import dataclass

import click

from robotframework_find_unused.common.cli import cli_hard_exit, pretty_kw_name
from robotframework_find_unused.common.const import KeywordData, KeywordFilterOption
from robotframework_find_unused.common.sort import sort_keywords_by_name

from .step.discover_files import cli_discover_file_paths
from .step.keyword_count_uses import cli_count_keyword_uses
from .step.keyword_definitions import cli_step_get_custom_keyword_definitions
from .step.keyword_filter import cli_filter_keywords
from .step.lib_keyword_definitions import cli_step_get_downloaded_lib_keywords
from .step.parse_files import cli_step_parse_files


@dataclass
class ReturnOptions:
    """
    Command line options for the 'returns' command
    """

    show_all_count: bool
    deprecated_keywords: KeywordFilterOption
    private_keywords: KeywordFilterOption
    library_keywords: KeywordFilterOption
    unused_keywords: KeywordFilterOption
    keyword_filter_glob: str | None
    verbose: int
    source_path: str


def cli_returns(options: ReturnOptions):
    """
    Entry point for the CLI command
    """
    file_paths = cli_discover_file_paths(options.source_path, verbose=options.verbose)
    if len(file_paths) == 0:
        return cli_hard_exit(options.verbose)

    files = cli_step_parse_files(
        file_paths,
        verbose=options.verbose,
    )

    keywords = cli_step_get_custom_keyword_definitions(
        files,
        verbose=options.verbose,
        enrich_py_keywords=True,
    )
    if len(keywords) == 0:
        return cli_hard_exit(options.verbose)

    downloaded_library_keywords = cli_step_get_downloaded_lib_keywords(
        file_paths,
        verbose=options.verbose,
        enrich_py_keywords=options.library_keywords != "exclude",
    )

    counted_keywords = cli_count_keyword_uses(
        file_paths,
        keywords,
        downloaded_library_keywords=downloaded_library_keywords,
        verbose=options.verbose,
    )

    if options.library_keywords != "exclude" and options.unused_keywords != "exclude":
        for lib in downloaded_library_keywords:
            for kw in lib.keywords:
                if kw in counted_keywords:
                    continue
                counted_keywords.append(kw)

    counted_keywords = cli_filter_keywords(
        counted_keywords,
        filter_deprecated=options.deprecated_keywords,
        filter_private=options.private_keywords,
        filter_library=options.library_keywords,
        filter_unused=options.unused_keywords,
        filter_returns="only",
        filter_glob=options.keyword_filter_glob,
    )
    _cli_log_results(counted_keywords, options)
    return _exit_code(counted_keywords)


def _cli_log_results(keywords: list[KeywordData], options: ReturnOptions) -> None:
    click.echo()

    if options.show_all_count:
        sorted_keywords = sort_keywords_by_name(keywords)
        sorted_keywords = sorted(sorted_keywords, key=lambda kw: kw.return_use_count)

        click.echo("return_use_count\tkeyword_name")
        for kw in sorted_keywords:
            click.echo("\t".join([str(kw.return_use_count), pretty_kw_name(kw)]))
    else:
        unused_returns = [kw for kw in keywords if kw.return_use_count == 0]
        unused_returns = sort_keywords_by_name(unused_returns)

        click.echo(f"Found {len(unused_returns)} unused keyword returns:")
        for kw in unused_returns:
            click.echo("  " + pretty_kw_name(kw))


def _exit_code(keywords: list[KeywordData]) -> int:
    unused_returns = [kw for kw in keywords if kw.return_use_count == 0]
    exit_code = len(unused_returns)
    return min(exit_code, 200)



"""
CLI frontend
"""

from .arguments import ArgumentsOptions, cli_arguments
from .keywords import KeywordOptions, cli_keywords
from .returns import ReturnOptions, cli_returns
from .variables import VariableOptions, cli_variables

__all__ = [
    "ArgumentsOptions",
    "KeywordOptions",
    "ReturnOptions",
    "VariableOptions",
    "cli_arguments",
    "cli_keywords",
    "cli_returns",
    "cli_variables",
]



from typing import cast

import click
from robot.libdocpkg.model import KeywordDoc, LibraryDoc

from robotframework_find_unused.common.cli import pretty_kw_name
from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    VERBOSE_NO,
    VERBOSE_SINGLE,
    KeywordData,
)
from robotframework_find_unused.common.convert import libdoc_keyword_to_keyword_data
from robotframework_find_unused.common.enrich_python_keywords import enrich_python_keyword_data


def cli_step_get_custom_keyword_definitions(
    files: list[LibraryDoc],
    *,
    verbose: int,
    enrich_py_keywords: bool = False,
):
    """
    Gather keyword definitions in the given scope with LibDoc and show progress
    """
    click.echo("Gathering custom keyword definitions...")

    keywords = _get_custom_keyword_definitions(
        files,
        enrich_py_keywords=enrich_py_keywords,
    )

    _log_keyword_stats(keywords, verbose)
    return keywords


def _get_custom_keyword_definitions(
    files: list[LibraryDoc],
    *,
    enrich_py_keywords: bool = False,
) -> list[KeywordData]:
    """
    Gather keyword definitions in the given scope with LibDoc

    Libdoc supports .robot, .resource, .py, and downloaded libs
    """
    keywords: list[KeywordData] = []
    for file in files:
        if file.type == "SUITE":
            file_type = "CUSTOM_SUITE"
        elif file.type == "LIBRARY":
            file_type = "CUSTOM_LIBRARY"
        elif file.type == "RESOURCE":
            file_type = "CUSTOM_RESOURCE"
        else:
            raise ValueError("Unexpected file type " + file.type)

        if file_type == "CUSTOM_LIBRARY" and enrich_py_keywords:
            enriched_keywords = enrich_python_keyword_data(file)
            for keyword in enriched_keywords:
                keywords.append(
                    libdoc_keyword_to_keyword_data(
                        keyword.doc,
                        file_type,
                        keyword.returns,
                    ),
                )
        else:
            # LibDoc provides all the data we need
            for keyword in cast(list[KeywordDoc], file.keywords):
                keywords.append(
                    libdoc_keyword_to_keyword_data(
                        keyword,
                        file_type,
                        # We don't care or will gather this later
                        keyword_returns=None,
                    ),
                )

    return keywords


def _log_keyword_stats(keywords: list[KeywordData], verbose: int) -> None:
    """
    Output details on the given keywords to the user
    """
    click.echo(
        (ERROR_MARKER if len(keywords) == 0 else DONE_MARKER)
        + f" Found {len(keywords)} custom keyword definitions",
    )

    if verbose == VERBOSE_NO:
        return

    kw_types: dict[str, list[str]] = {}
    for kw in keywords:
        if kw.type not in kw_types:
            kw_types[kw.type] = []
        kw_types[kw.type].append(pretty_kw_name(kw))

    for kw_type, kw_names in sorted(kw_types.items(), key=lambda x: len(x[1]), reverse=True):
        click.echo(f"{INDENT}{len(kw_names)} keywords of type {kw_type}")

        if verbose == VERBOSE_SINGLE:
            continue
        for name in kw_names:
            click.echo(f"{INDENT}{INDENT}{name}")



from .const import VariableData
from .normalize import normalize_variable_name

supported_builtin_vars = {
    "true": "True",
    "false": "False",
    "none": "None",
    "empty": "",
    "space": " ",
}


def get_variables_in_string(input_string: str) -> list[str]:
    """
    Return the Robot variables in a string
    """
    variables: list[str] = []

    string = input_string
    while len(string) > 0:
        string = _find_variable_start(string)
        (string, variable) = _find_variable_end(string)

        if variable:
            variables.append(variable)

    return variables


def _find_variable_start(string: str) -> str:
    while len(string) > 0 and string[0] not in "$&@%":
        string = string[1:]
    return string


def _find_variable_end(string: str) -> tuple[str, str | None]:
    variable = ""
    depth = 0
    while len(string) > 0:
        char = string[0]
        if char == "{":
            depth += 1
        if char == "}":
            depth -= 1

        variable += char
        string = string[1:]

        if char == "}" and depth == 0:
            break

    return (string, variable or None)


def resolve_variables(
    robot_input: str,
    variables: dict[str, VariableData],
) -> tuple[str, list[str]]:
    """
    Resolve variables in the given string.

    Only resolves simple cases of builtin vars and given vars.
    """
    used_vars = get_variables_in_string(robot_input)

    replaced_vars: list[str] = []
    for var in used_vars:
        var_normalized = normalize_variable_name(var)
        val = None
        if var_normalized not in variables:
            try:
                val = _get_value_of_builtin_var(var_normalized)
            except ValueError:
                # Not known: Not simple. Don't try.
                continue

        if val is None:
            val = tuple(variables[var_normalized].value)
            if len(val) == 0:
                val = ""
            elif len(val) == 1:
                val = val[0]
            else:
                # Not single-line scalar: Not simple. Don't try.
                continue

        if val is not None:
            robot_input = robot_input.replace(var, val)
            replaced_vars.append(var_normalized)

    return (robot_input, replaced_vars)


def _get_value_of_builtin_var(normalized_name: str) -> str:
    if normalized_name in supported_builtin_vars:
        return supported_builtin_vars[normalized_name]

    stripped_var = normalized_name.removeprefix("{").removesuffix("}")
    try:
        float(stripped_var)
    except ValueError:
        pass
    else:
        # Is a number, not a variable name.
        return stripped_var

    msg = f"Can't get value of unsupported builtin variable '${normalized_name}'"
    raise ValueError(msg)


def resolve_variable_name(
    var_name: str,
    variables: dict[str, VariableData],
) -> tuple[str, list[str]]:
    """
    Resolve variable name.

    Returns tuple of (resolved_var_name, used_vars)
    """
    if not ("${" in var_name or "@{" in var_name or "&{" in var_name or "%{" in var_name):
        return (var_name, [])

    (resolved, used_vars) = resolve_variables(var_name, variables)
    if var_name == resolved:
        return (var_name, [])

    resolved_var = normalize_variable_name(resolved, strip_decoration=False)
    if resolved_var in variables:
        return (normalize_variable_name(resolved), used_vars)

    (recursed_resolved, recursed_used_vars) = resolve_variable_name(resolved, variables)
    return (recursed_resolved, used_vars + recursed_used_vars)
